/* Copyright (C) 2007, 2008, 2009 Stefan Frenz and Patrick Schmidt
 * 
 * This file is part of SJC, the Small Java Compiler written by Stefan Frenz.
 * 
 * SJC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * SJC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with SJC. If not, see <http://www.gnu.org/licenses/>.
 */

package symbols;

import osio.TextPrinter;
import compbase.*;

/**
 * BootStrap: generate output for all compiler relevant symbol information
 * 
 * @author S. Frenz, P. Schmidt
 * @version 091105 added option noSLHI to avoid source line hint integration even if existing, moved genIntArray to Context
 *  version 091104 added support for lineInCodeOffset for methods
 *  version 091102 added filling of outerUnit and innerUnits in Units
 *  version 091021 adopted changed modifier declarations
 *  version 091001 adopted changed memory interface
 *  version 090718 adopted move of modifier flags from Vrbl to AccVar
 *  version 090609 fixed putRef instead of putInt in genVrbl
 *  version 090207 added copyright notice
 *  version 080614 adopted changed Unit.searchVariable
 *  version 080327 fixed filling of method symHint
 *  version 080115 adopted changes in SymbolInformer.generateSymbols, using SymbolInformer.checkVrbl
 *  version 080114 removed checkParameter
 *  version 070920 stricter test of root variable declaration, fixed root variable setting in dynaMem mode
 *  version 070816 removed osio from compiler-exlusion list, added parameter to avoid null-unitDest
 *  version 070801 added Unit.offset*
 *  version 070731 adopted QualIDList, adopted renaming of id to name
 *  version 070730 fixed option "noCompilerPacks" to include packages but to exclude units
 *  version 070729 added option "noCompilerPacks" to reduce size of added symbol information
 *  version 070729 removed no longer needed prepareUnits, added type check for root-variable, fixed some bugs
 *  version 070727 initial version
 */

public class BootStrapSymbols extends SymbolInformer {
  private boolean noCompilerPacks;
  private Unit packUnit, unitUnit, mthdUnit, vrblUnit, paramUnit, typeRefUnit, qualIDUnit, qualIDListUnit;
  private Unit stringListUnit;
  private int packNamePos, packOuterPos, packSubPacksPos, packNextPackPos, packUnitsPos;
  private int unitNamePos, unitExtsIDPos, unitExtsImplIDListPos, unitModifierPos;
  private int unitNextUnitPos, unitOuterUnitPos, unitInnerUnitsPos;
  private int unitPackPos, unitMthdsPos, unitVarsPos, unitInlArrPos, unitOutputLocationPos;
  private int unitExplConstrPos, unitExplStdConstrPos, unitInitStatPos, unitInitDynaPos, unitStatRelocTableEntriesPos;
  private int unitStatScalarTableSizePos, unitClssRelocTableEntriesPos, unitClssScalarTableSizePos;
  private int unitInstRelocTableEntriesPos, unitInstScalarTableSizePos, unitInstIndirScalarTableSizePos;
  private int unitOffsetsAssignedPos, unitOutputGeneratedPos;
  private int mthdRetTypePos, mthdOwnerPos, mthdMarkerPos, mthdRelOffPos, mthdIsConstructorPos;
  private int mthdOvldMthdPos, mthdRedirectPos, mthdOutputLocationPos, mthdParamPos;
  private int mthdNamePos, mthdModifierPos, mthdNextMthdPos, mthdLineInCodeOffsetPos;
  private int vrblOwnerPos, vrblModifierPos, vrblNextVrblPos, vrblNamePos, vrblTypePos;
  private int vrblMinSizePos, vrblRelOffPos, vrblLocationPos;
  private int paramNextParamPos, paramNamePos, paramTypePos, paramMinSizePos, paramRelOffPos;
  private int paramLocationPos;
  private int typeRefBaseTypePos, typeRefArrDimPos, typeRefInlinedArraySizePos, typeRefQidPos;
  private int qualIDNamePos, qualIDTypePos, qualIDPackDestPos, qualIDUnitDestPos;
  private int qualIDListQidPos, qualIDListNextPos;
  private int stringListStrPos, stringListNextPos, stringListTablePosPos;
  private TypeRefPool typeRefPool;
  private boolean noSLHI;
  
  private final static String NOSLHI = "skipping source line information in BootStrapSymbols (";
  
  public static void printValidParameters(TextPrinter v) {
    v.println("   noCompilerPacks   exclude packages of compiler from symbol information");
    v.println("   noSLHI            exclude source line hint integration information");
  }
  
  public boolean setParameter(String parm, TextPrinter v) {
    if (parm.equals("noCompilerPacks")) {
      noCompilerPacks=true;
      return true;
    }
    if (parm.equals("noSLHI")) {
      noSLHI=true;
      return true;
    }
    v.println("BootStrapSymbols accepts only the following parameters:");
    printValidParameters(v);
    return false;
  }

  public boolean generateSymbols(UnitList newUnits, Context ictx) {
    Object rootLoc;
    Unit myUnit;
    Vrbl rootVar;
    StringList list;
    Pack compbasePack;
    
    ctx=ictx;
    typeRefPool=new TypeRefPool();
    //check basics
    list=new StringList("symbols");
    list.next=new StringList("BootStrap");
    if ((myUnit=ctx.root.searchUnit(list))==null
        || (myUnit.modifier&(Modifier.M_ABSTR|Modifier.M_INDIR|Modifier.M_STRUCT))!=0) {
      ctx.out.println("unit symbols.BootStrap required for BootStrapSymbols (non-abstract, non-indir, non-struct)");
      return false;
    }
    if ((compbasePack=ctx.root.searchSubPackage("compbase"))==null
        || (packUnit=compbasePack.searchUnit("Pack"))==null
        || (unitUnit=compbasePack.searchUnit("UnitDummy"))==null
        || (mthdUnit=compbasePack.searchUnit("MthdDummy"))==null
        || (vrblUnit=compbasePack.searchUnit("Vrbl"))==null
        || (paramUnit=compbasePack.searchUnit("Param"))==null
        || (typeRefUnit=compbasePack.searchUnit("TypeRef"))==null
        || (qualIDUnit=compbasePack.searchUnit("QualID"))==null
        || (qualIDListUnit=compbasePack.searchUnit("QualIDList"))==null
        || (stringListUnit=compbasePack.searchUnit("StringList"))==null) {
      ctx.out.println("required packages or units not found");
      return false;
    }
    if ((rootVar=myUnit.searchVariable("root", ctx))==null
        || (rootVar.modifier&(Modifier.M_FIN|Modifier.M_STAT|Modifier.M_STRUCT))!=Modifier.M_STAT
        || rootVar.type.baseType!=TypeRef.T_QID || rootVar.type.qid.unitDest!=packUnit
        || rootVar.type.arrDim!=0) {
      ctx.out.println("variable \"static Pack root\" needed in symbols.BootStrap");
      return false;
    }
    if ((packNamePos=checkVrbl(packUnit, "name"))==AccVar.INV_RELOFF
        || (packOuterPos=checkVrbl(packUnit, "outer"))==AccVar.INV_RELOFF
        || (packSubPacksPos=checkVrbl(packUnit, "subPacks"))==AccVar.INV_RELOFF
        || (packNextPackPos=checkVrbl(packUnit, "nextPack"))==AccVar.INV_RELOFF
        || (packUnitsPos=checkVrbl(packUnit, "units"))==AccVar.INV_RELOFF
        || (unitNamePos=checkVrbl(unitUnit, "name"))==AccVar.INV_RELOFF
        || (unitExtsIDPos=checkVrbl(unitUnit, "extsID"))==AccVar.INV_RELOFF
        || (unitExtsImplIDListPos=checkVrbl(unitUnit, "extsImplIDList"))==AccVar.INV_RELOFF
        || (unitModifierPos=checkVrbl(unitUnit, "modifier"))==AccVar.INV_RELOFF
        || (unitNextUnitPos=checkVrbl(unitUnit, "nextUnit"))==AccVar.INV_RELOFF
        || (unitOuterUnitPos=checkVrbl(unitUnit, "outerUnit"))==AccVar.INV_RELOFF
        || (unitInnerUnitsPos=checkVrbl(unitUnit, "innerUnits"))==AccVar.INV_RELOFF
        || (unitPackPos=checkVrbl(unitUnit, "pack"))==AccVar.INV_RELOFF
        || (unitMthdsPos=checkVrbl(unitUnit, "mthds"))==AccVar.INV_RELOFF
        || (unitVarsPos=checkVrbl(unitUnit, "vars"))==AccVar.INV_RELOFF
        || (unitInlArrPos=checkVrbl(unitUnit, "inlArr"))==AccVar.INV_RELOFF
        || (unitOutputLocationPos=checkVrbl(unitUnit, "outputLocation"))==AccVar.INV_RELOFF
        || (unitExplConstrPos=checkVrbl(unitUnit, "explConstr"))==AccVar.INV_RELOFF
        || (unitExplStdConstrPos=checkVrbl(unitUnit, "explStdConstr"))==AccVar.INV_RELOFF
        || (unitInitStatPos=checkVrbl(unitUnit, "initStat"))==AccVar.INV_RELOFF
        || (unitInitDynaPos=checkVrbl(unitUnit, "initDyna"))==AccVar.INV_RELOFF
        || (unitStatRelocTableEntriesPos=checkVrbl(unitUnit, "statRelocTableEntries"))==AccVar.INV_RELOFF
        || (unitStatScalarTableSizePos=checkVrbl(unitUnit, "statScalarTableSize"))==AccVar.INV_RELOFF
        || (unitClssRelocTableEntriesPos=checkVrbl(unitUnit, "clssRelocTableEntries"))==AccVar.INV_RELOFF
        || (unitClssScalarTableSizePos=checkVrbl(unitUnit, "clssScalarTableSize"))==AccVar.INV_RELOFF
        || (unitInstRelocTableEntriesPos=checkVrbl(unitUnit, "instRelocTableEntries"))==AccVar.INV_RELOFF
        || (unitInstScalarTableSizePos=checkVrbl(unitUnit, "instScalarTableSize"))==AccVar.INV_RELOFF
        || (unitInstIndirScalarTableSizePos=checkVrbl(unitUnit, "instIndirScalarTableSize"))==AccVar.INV_RELOFF
        || (unitOffsetsAssignedPos=checkVrbl(unitUnit, "offsetsAssigned"))==AccVar.INV_RELOFF
        || (unitOutputGeneratedPos=checkVrbl(unitUnit, "outputGenerated"))==AccVar.INV_RELOFF
        || (mthdNamePos=checkVrbl(mthdUnit, "name"))==AccVar.INV_RELOFF
        || (mthdRetTypePos=checkVrbl(mthdUnit, "retType"))==AccVar.INV_RELOFF
        || (mthdNextMthdPos=checkVrbl(mthdUnit, "nextMthd"))==AccVar.INV_RELOFF
        || (mthdOwnerPos=checkVrbl(mthdUnit, "owner"))==AccVar.INV_RELOFF
        || (mthdModifierPos=checkVrbl(mthdUnit, "modifier"))==AccVar.INV_RELOFF
        || (mthdMarkerPos=checkVrbl(mthdUnit, "marker"))==AccVar.INV_RELOFF
        || (mthdRelOffPos=checkVrbl(mthdUnit, "relOff"))==AccVar.INV_RELOFF
        || (mthdIsConstructorPos=checkVrbl(mthdUnit, "isConstructor"))==AccVar.INV_RELOFF
        || (mthdOvldMthdPos=checkVrbl(mthdUnit, "ovldMthd"))==AccVar.INV_RELOFF
        || (mthdRedirectPos=checkVrbl(mthdUnit, "redirect"))==AccVar.INV_RELOFF
        || (mthdOutputLocationPos=checkVrbl(mthdUnit, "outputLocation"))==AccVar.INV_RELOFF
        || (mthdParamPos=checkVrbl(mthdUnit, "param"))==AccVar.INV_RELOFF
        || (vrblOwnerPos=checkVrbl(vrblUnit, "owner"))==AccVar.INV_RELOFF
        || (vrblModifierPos=checkVrbl(vrblUnit, "modifier"))==AccVar.INV_RELOFF
        || (vrblNextVrblPos=checkVrbl(vrblUnit, "nextVrbl"))==AccVar.INV_RELOFF
        || (vrblNamePos=checkVrbl(vrblUnit, "name"))==AccVar.INV_RELOFF
        || (vrblTypePos=checkVrbl(vrblUnit, "type"))==AccVar.INV_RELOFF
        || (vrblMinSizePos=checkVrbl(vrblUnit, "minSize"))==AccVar.INV_RELOFF
        || (vrblRelOffPos=checkVrbl(vrblUnit, "relOff"))==AccVar.INV_RELOFF
        || (vrblLocationPos=checkVrbl(vrblUnit, "location"))==AccVar.INV_RELOFF
        || (paramNextParamPos=checkVrbl(paramUnit, "nextParam"))==AccVar.INV_RELOFF
        || (paramNamePos=checkVrbl(paramUnit, "name"))==AccVar.INV_RELOFF
        || (paramTypePos=checkVrbl(paramUnit, "type"))==AccVar.INV_RELOFF
        || (paramMinSizePos=checkVrbl(paramUnit, "minSize"))==AccVar.INV_RELOFF
        || (paramRelOffPos=checkVrbl(paramUnit, "relOff"))==AccVar.INV_RELOFF
        || (paramLocationPos=checkVrbl(paramUnit, "location"))==AccVar.INV_RELOFF
        || (typeRefBaseTypePos=checkVrbl(typeRefUnit, "baseType"))==AccVar.INV_RELOFF
        || (typeRefArrDimPos=checkVrbl(typeRefUnit, "arrDim"))==AccVar.INV_RELOFF
        || (typeRefInlinedArraySizePos=checkVrbl(typeRefUnit, "inlinedArraySize"))==AccVar.INV_RELOFF
        || (typeRefQidPos=checkVrbl(typeRefUnit, "qid"))==AccVar.INV_RELOFF
        || (qualIDNamePos=checkVrbl(qualIDUnit, "name"))==AccVar.INV_RELOFF
        || (qualIDTypePos=checkVrbl(qualIDUnit, "type"))==AccVar.INV_RELOFF
        || (qualIDPackDestPos=checkVrbl(qualIDUnit, "packDest"))==AccVar.INV_RELOFF
        || (qualIDUnitDestPos=checkVrbl(qualIDUnit, "unitDest"))==AccVar.INV_RELOFF
        || (qualIDListQidPos=checkVrbl(qualIDListUnit, "qid"))==AccVar.INV_RELOFF
        || (qualIDListNextPos=checkVrbl(qualIDListUnit, "nextQualID"))==AccVar.INV_RELOFF
        || (stringListStrPos=checkVrbl(stringListUnit, "str"))==AccVar.INV_RELOFF
        || (stringListNextPos=checkVrbl(stringListUnit, "next"))==AccVar.INV_RELOFF
        || (stringListTablePosPos=checkVrbl(stringListUnit, "tablePos"))==AccVar.INV_RELOFF) {
      return false;
    }
    if (noSLHI) {
      ctx.out.print(NOSLHI);
      ctx.out.println("explicit off-option given)");
      mthdLineInCodeOffsetPos=AccVar.INV_RELOFF;
    }
    else if ((mthdLineInCodeOffsetPos=checkVrbl(mthdUnit, "lineInCodeOffset"))==AccVar.INV_RELOFF) { 
      ctx.out.print(NOSLHI);
      ctx.out.println("missing variable)");
    }
    //create objects for packages, units, methods and variables; enter root in static var
    if ((rootLoc=genPackage(ctx.root))==null) return false;
    ctx.arch.putRef(myUnit.outputLocation, rootVar.relOff, rootLoc, 0);
    //everything done
    return !ctx.err;
  }
  
  private Object alloc(Unit unit) {
    Object loc;
    if (ctx.err) return null; //no error printing if an error occurred already 
    if ((loc=ctx.mem.allocate(unit.instScalarTableSize, unit.instIndirScalarTableSize,
        unit.instRelocTableEntries, unit.outputLocation))==null) {
      ctx.out.print("error allocating ");
      ctx.out.print(unit.name);
      ctx.out.println("-object for BootStrapSymbols");
      ctx.err=true;
    }
    return loc;
  }
  
  private boolean isCompilerPack(Pack pack) {
    String n;
    
    return pack!=null && (n=pack.name)!=null && (n.equals("backend")
        || n.equals("compbase")  || n.equals("compression")
        || n.equals("emulation") || n.equals("frontend")
        || n.equals("memory") || n.equals("output") || n.equals("symbols"));
  }
  
  private Object genPackage(Pack pack) {
    Object packObj=null;
    
    if (pack!=null) {
      if (pack.symHint!=null) return pack.symHint;
      if ((packObj=alloc(packUnit))==null) return null;
      pack.symHint=packObj;
      ctx.arch.putRef(packObj, packNamePos, ctx.allocateString(pack.name), 0);
      ctx.arch.putRef(packObj, packOuterPos, genPackage(pack.outer), 0);
      if (!noCompilerPacks || !isCompilerPack(pack)) {
        ctx.arch.putRef(packObj, packUnitsPos, genUnit(pack.units, false), 0);
        ctx.arch.putRef(packObj, packSubPacksPos, genPackage(pack.subPacks), 0);
      }
      ctx.arch.putRef(packObj, packNextPackPos, genPackage(pack.nextPack), 0);
    }
    return packObj;
  }
  
  private Object genUnit(Unit unit, boolean avoidNull) {
    Object unitObj=null;
    
    if (unit!=null) {
      if (noCompilerPacks && unit.pack!=null && isCompilerPack(unit.pack.packDest))
        return avoidNull ? genUnit(ctx.langRoot.unitDest, false) : null;
      if (unit.symHint!=null) return unit.symHint;
      if ((unitObj=alloc(unitUnit))==null) return null;
      unit.symHint=unitObj;
      ctx.arch.putRef(unitObj, unitNamePos, ctx.allocateString(unit.name), 0);
      ctx.arch.putRef(unitObj, unitPackPos, genQualID(unit.pack), 0);
      ctx.arch.putRef(unitObj, unitMthdsPos, genMthd(unit.mthds), 0);
      ctx.mem.putInt(unitObj, unitModifierPos, unit.modifier);
      ctx.arch.putRef(unitObj, unitExtsIDPos, genQualID(unit.extsID), 0);
      ctx.arch.putRef(unitObj, unitExtsImplIDListPos, genQualIDList(unit.extsImplIDList), 0);
      ctx.arch.putRef(unitObj, unitOutputLocationPos, unit.outputLocation, 0);
      ctx.mem.putByte(unitObj, unitExplConstrPos, unit.explConstr ? (byte)1 : (byte)0);
      ctx.mem.putByte(unitObj, unitExplStdConstrPos, unit.explStdConstr ? (byte)1 : (byte)0);
      ctx.arch.putRef(unitObj, unitVarsPos, genVrbl(unit.vars), 0);
      ctx.arch.putRef(unitObj, unitInlArrPos, genVrbl(unit.inlArr), 0);
      ctx.arch.putRef(unitObj, unitInitStatPos, genMthd(unit.initStat), 0);
      ctx.arch.putRef(unitObj, unitInitDynaPos, genMthd(unit.initDyna), 0);
      ctx.arch.putRef(unitObj, unitNextUnitPos, genUnit(unit.nextUnit, avoidNull), 0);
      ctx.arch.putRef(unitObj, unitOuterUnitPos, genUnit(unit.outerUnit, avoidNull), 0);
      ctx.arch.putRef(unitObj, unitInnerUnitsPos, genUnit(unit.innerUnits, avoidNull), 0);
      ctx.mem.putInt(unitObj, unitStatRelocTableEntriesPos, unit.statRelocTableEntries);
      ctx.mem.putInt(unitObj, unitStatScalarTableSizePos, unit.statScalarTableSize);
      ctx.mem.putInt(unitObj, unitClssRelocTableEntriesPos, unit.clssRelocTableEntries);
      ctx.mem.putInt(unitObj, unitClssScalarTableSizePos, unit.clssScalarTableSize);
      ctx.mem.putInt(unitObj, unitInstRelocTableEntriesPos, unit.instRelocTableEntries);
      ctx.mem.putInt(unitObj, unitInstScalarTableSizePos, unit.instScalarTableSize);
      ctx.mem.putInt(unitObj, unitInstIndirScalarTableSizePos, unit.instIndirScalarTableSize);
      ctx.mem.putByte(unitObj, unitOffsetsAssignedPos, unit.offsetsAssigned ? (byte)1 : (byte)0);
      ctx.mem.putByte(unitObj, unitOutputGeneratedPos, unit.outputGenerated ? (byte)1 : (byte)0);
    }
    return unitObj;
  }
  
  private Object genMthd(Mthd mthd) {
    Object mthdObj=null;
    
    if (mthd!=null) {
      if (mthd.symHint!=null) return mthd.symHint;
      if ((mthdObj=alloc(mthdUnit))==null) return null;
      mthd.symHint=mthdObj;
      ctx.arch.putRef(mthdObj, mthdNextMthdPos, genMthd(mthd.nextMthd), 0);
      ctx.arch.putRef(mthdObj, mthdNamePos, ctx.allocateString(mthd.name), 0);
      ctx.arch.putRef(mthdObj, mthdOvldMthdPos, genMthd(mthd.ovldMthd), 0);
      ctx.mem.putInt(mthdObj, mthdModifierPos, mthd.modifier);
      ctx.arch.putRef(mthdObj, mthdRetTypePos, genTypeRef(mthd.retType), 0);
      ctx.arch.putRef(mthdObj, mthdOwnerPos, genUnit(mthd.owner, true), 0);
      ctx.mem.putInt(mthdObj, mthdMarkerPos, mthd.marker);
      ctx.mem.putInt(mthdObj, mthdRelOffPos, mthd.relOff);
      ctx.mem.putByte(mthdObj, mthdIsConstructorPos, mthd.isConstructor ? (byte)1 : (byte)0);
      ctx.arch.putRef(mthdObj, mthdRedirectPos, genMthd(mthd.redirect), 0);
      ctx.arch.putRef(mthdObj, mthdOutputLocationPos, mthd.outputLocation, 0);
      ctx.arch.putRef(mthdObj, mthdParamPos, genParam(mthd.param), 0);
      if (mthdLineInCodeOffsetPos!=AccVar.INV_RELOFF) {
        ctx.arch.putRef(mthdObj, mthdLineInCodeOffsetPos, ctx.allocateIntArray(mthd.lineInCodeOffset), 0);
      }
    }
    return mthdObj;
  }
  
  private Object genTypeRef(TypeRef typeRef) {
    Object typeRefObj=null;
    
    if (typeRef!=null) {
      if (typeRef.symHint!=null) return typeRef.symHint;
      if ((typeRefObj=typeRefPool.findMatchingTypeRef(typeRef))!=null) return typeRefObj;
      if ((typeRefObj=alloc(typeRefUnit))==null) return null;
      typeRef.symHint=typeRefObj;
      ctx.mem.putInt(typeRefObj, typeRefBaseTypePos, typeRef.baseType);
      ctx.mem.putInt(typeRefObj, typeRefArrDimPos, typeRef.arrDim);
      ctx.mem.putInt(typeRefObj, typeRefInlinedArraySizePos, typeRef.inlinedArraySize);
      ctx.arch.putRef(typeRefObj, typeRefQidPos, genQualID(typeRef.qid), 0);
    }
    return typeRefObj;
  }
  
  private Object genParam(Param param) {
    Object paramObj=null;
    
    if (param!=null) {
      if ((paramObj=alloc(paramUnit))==null) return null;
      ctx.arch.putRef(paramObj, paramNextParamPos, genParam(param.nextParam), 0);
      ctx.arch.putRef(paramObj, paramNamePos, ctx.allocateString(param.name), 0);
      ctx.arch.putRef(paramObj, paramTypePos, genTypeRef(param.type), 0);
      ctx.mem.putInt(paramObj, paramMinSizePos, param.minSize);
      ctx.mem.putInt(paramObj, paramRelOffPos, param.relOff);
      ctx.mem.putInt(paramObj, paramLocationPos, param.location);
    }
    return paramObj;
  }
  
  private Object genQualID(QualID qualID) {
    Object qualIDObj=null;
    
    if (qualID!=null) {
      if ((qualIDObj=alloc(qualIDUnit))==null) return null;
      ctx.mem.putInt(qualIDObj, qualIDTypePos, qualID.type);
      ctx.arch.putRef(qualIDObj, qualIDNamePos, genStringList(qualID.name), 0);
      ctx.arch.putRef(qualIDObj, qualIDPackDestPos, genPackage(qualID.packDest), 0);
      ctx.arch.putRef(qualIDObj, qualIDUnitDestPos, genUnit(qualID.unitDest, true), 0);
    }
    return qualIDObj;
  }
  
  private Object genQualIDList(QualIDList qualIDList) {
    Object qualIDListAddr=null;
    
    if (qualIDList!=null) {
      if ((qualIDListAddr=alloc(qualIDListUnit))==null) return null;
      ctx.arch.putRef(qualIDListAddr, qualIDListQidPos, genQualID(qualIDList.qid), 0);
      ctx.arch.putRef(qualIDListAddr, qualIDListNextPos, genQualIDList(qualIDList.nextQualID), 0);
    }
    return qualIDListAddr;
  }
  
  private Object genVrbl(Vrbl vrbl) {
    Object vrblObj=null;
    
    if (vrbl!=null) {
      if (vrbl.symHint!=null) return vrbl.symHint;
      if ((vrblObj=alloc(vrblUnit))==null) return null;
      ctx.arch.putRef(vrblObj, vrblOwnerPos, genUnit(vrbl.owner, true), 0);
      ctx.mem.putInt(vrblObj, vrblModifierPos, vrbl.modifier);
      ctx.arch.putRef(vrblObj, vrblNextVrblPos, genVrbl(vrbl.nextVrbl), 0);
      ctx.arch.putRef(vrblObj, vrblNamePos, ctx.allocateString(vrbl.name), 0);
      ctx.arch.putRef(vrblObj, vrblTypePos, genTypeRef(vrbl.type), 0);
      ctx.mem.putInt(vrblObj, vrblMinSizePos, vrbl.minSize);
      ctx.mem.putInt(vrblObj, vrblRelOffPos, vrbl.relOff);
      ctx.mem.putInt(vrblObj, vrblLocationPos, vrbl.location);
    }
    return vrblObj;
  }
  
  private Object genStringList(StringList stringList) {
    Object stringListAddr=null;
    
    if (stringList!=null) {
      if ((stringListAddr=alloc(stringListUnit))==null) return null;
      ctx.arch.putRef(stringListAddr, stringListStrPos, ctx.allocateString(stringList.str), 0);
      ctx.arch.putRef(stringListAddr, stringListNextPos, genStringList(stringList.next), 0);
      ctx.mem.putInt(stringListAddr, stringListTablePosPos, stringList.tablePos);
    }
    return stringListAddr;
  }
}
